package ca.yorku.cse.mack.sak;

import java.util.Arrays;

/**
 * SAKStats - a program to provide stats and other services related to the
 * scanning ambiguous keyboard (SAK).
 *
 * Android version.
 *
 * @author Scott MacKenzie, 2008-2016
 */

public class SAKStats
{
    final static int OLPS = 0;
    final static int MLPS = 1;
    final static int MLPK = 2;
    final static int OW = 3;

    // ----------------------------------------------------------------------
    // return a scan sequence for the specified word using the specified mode
    // ----------------------------------------------------------------------
    public static String getScanSequence(int mode, String word, EncodedWord[] ew)
    {
        if (mode == SAKStats.OLPS)
            return getOLPSScanSequence(word.toLowerCase(), ew);
        else if (mode == SAKStats.MLPS)
            return getMLPSScanSequence(word.toLowerCase(), ew);
        else if (mode == SAKStats.MLPK)
            return getMLPKScanSequence(word.toLowerCase(), ew);
        else if (mode == SAKStats.OW)
            return getOWScanSequence(word.toLowerCase(), ew);
        else
            System.out.println("getScanSequence error: no or invalid mode argument!");
        return "?";
    }

    //-------------------------------------------------------------------------------
    // OLPS = One Letter Per Scan scanning method
    private static String getOLPSScanSequence(String word, EncodedWord[] ew)
    {
        int[] letterPosition = getLetterPosition(ew);
        String[] keyboard = getKeyboard(letterPosition);
        String keystrokes = getKeystrokes(word, letterPosition);
        String scanString = "";
        for (int j = 0; j < word.length(); ++j)
        {
            char c = word.charAt(j);
            int p = letterPosition[c - 'a'];
            for (int k = 0; k < p; ++k)
                scanString += ".";
            scanString += c;
            for (int k = 0; k < keyboard.length - p; ++k)
                scanString += ".";
        }
        scanString = scanString.substring(0, scanString.length() - 1); // drop last "."
        scanString += "S";
        String[] query = getCandidates(keystrokes, ew);
        if (query == null)
            return null;
        for (String q : query)
            if (word.equals(q))
                break;
            else
                scanString += ".";
        scanString += "W";
        return scanString;
    }

    //-------------------------------------------------------------------------------
    // MLPS = Multiple Letter Per Scan scanning method
    private static String getMLPSScanSequence(String word, EncodedWord[] ew)
    {
        int[] letterPosition = getLetterPosition(ew);
        String[] keyboard = getKeyboard(letterPosition);
        String keystrokes = getKeystrokes(word, letterPosition);
        int dist[][] = getDistanceMatrix(letterPosition);
        String scanString;

        char c1 = word.charAt(0);
        char c2;
        scanString = "";
        for (int j = 0; j < letterPosition[c1 - 'a']; ++j)
            scanString += ".";
        scanString += c1;

        for (int j = 1; j < word.length(); ++j)
        {
            c1 = word.charAt(j - 1);
            c2 = word.charAt(j);
            int idx1 = c1 - 'a';
            int idx2 = c2 - 'a';
            int d = dist[idx1][idx2];
            for (int k = 0; k < d - 1; ++k)
                scanString += ".";
            scanString += c2;
        }
        c2 = word.charAt(word.length() - 1); // last letter in word
        int n = letterPosition[c2 - 'a'];
        for (int k = 0; k < keyboard.length - n - 1; ++k)
            scanString += ".";
        scanString += "S";
        String[] query = getCandidates(keystrokes, ew);
        if (query == null)
            return null;
        for (String q : query)
            if (word.equals(q))
                break;
            else
                scanString += ".";
        scanString += "W";
        return scanString;
    }

    //-------------------------------------------------------------------------------
    // MLPK = Multiple Letter Per Key scanning method
    private static String getMLPKScanSequence(String word, EncodedWord[] ew)
    {
        int[] letterPosition = getLetterPosition(ew);
        String[] keyboard = getKeyboard(letterPosition);
        String keystrokes = getKeystrokes(word, letterPosition);
        int dist[][] = getDistanceMatrix(letterPosition);
        String scanString;

        String reducedWord = "";
        char c1, c2;

        // Reduce the word!  Where multiple letters are on the same key,
        // only show the 1st letter of the group
        for (int j = 0; j < keystrokes.length(); ++j)
        {
            reducedWord += word.charAt(j);
            while (j < keystrokes.length() - 1 &&
                    keystrokes.charAt(j) == keystrokes.charAt(j + 1))
                ++j;
        }

        // now proceed with "reduced word", as per MLPS method
        c1 = reducedWord.charAt(0);
        scanString = "";
        for (int j = 0; j < letterPosition[c1 - 'a']; ++j)
            scanString += ".";
        scanString += c1;

        for (int j = 1; j < reducedWord.length(); ++j)
        {
            c1 = reducedWord.charAt(j - 1);
            c2 = reducedWord.charAt(j);
            int idx1 = c1 - 'a';
            int idx2 = c2 - 'a';
            int d = dist[idx1][idx2];
            for (int k = 0; k < d - 1; ++k)
                scanString += ".";
            scanString += c2;
        }
        c2 = reducedWord.charAt(reducedWord.length() - 1); // last letter in word
        int n = letterPosition[c2 - 'a'];
        for (int k = 0; k < keyboard.length - n - 1; ++k)
            scanString += ".";
        scanString += "S";
        String[] query = getCandidates(keystrokes, ew);
        if (query == null)
            return null;
        for (String q : query)
            if (word.equals(q))
                break;
            else
                scanString += ".";
        scanString += "W";
        return scanString;
    }

    //-------------------------------------------------------------------------------
    // OW = Optimized Word scanning method
    private static String getOWScanSequence(String word, EncodedWord[] ew)
    {
        String scanString;
        int[] letterPosition;
        String[] keyboard;
        letterPosition = getLetterPosition(ew);
        keyboard = getKeyboard(letterPosition);
        String keystrokes = getKeystrokes(word, letterPosition);

        // build an array showing position of candidate word in list for
        // each word stem (e.g., "hello" --> h, he, hel, hell, hello)
        int[] candidateIndex = new int[word.length()];
        for (int j = 0; j < word.length(); ++j)
        {
            String keystrokeStem = keystrokes.substring(0, j + 1);
            String[] tryQuery = getCandidates(keystrokeStem, ew);
            if (tryQuery == null)
                return null;
            candidateIndex[j] = -1; // assume word not in candidate list
            for (int k = 0; k < tryQuery.length; ++k)
                if (word.equals(tryQuery[k]))
                {
                    candidateIndex[j] = k;
                    break;
                }
        }

        // use MLPK scan string as the starting point
        scanString = getMLPKScanSequence(word, ew);
        String optimizedScanString = "";
        String wordStem = "";
        for (int i = 0; scanString != null && i < scanString.length(); ++i)
        {
            char c = scanString.charAt(i);
            if (Character.isLowerCase(c))
            {
                String scanStem = scanString.substring(0, i + 1);
                wordStem += c;
                while (wordStem.length() < word.length())
                {
                    char next = word.charAt(wordStem.length());
                    if (letterPosition[c - 'a'] == letterPosition[next - 'a'])
                        wordStem += next;
                    else
                        break;
                }
                int p1 = scanStem.length() + (keyboard.length - letterPosition[c - 'a']);
                int p2 = candidateIndex[wordStem.length() - 1];
                if (p2 < 0) p2 = 99999;
                if ((p1 + p2) < scanString.length())
                {
                    optimizedScanString = scanStem;
                    for (int j = 0; j < keyboard.length - letterPosition[c - 'a'] - 1; ++j)
                        optimizedScanString += ".";
                    optimizedScanString += "S";
                    for (int j = 0; j < candidateIndex[wordStem.length() - 1]; ++j)
                        optimizedScanString += ".";
                    optimizedScanString += "W";
                    break;
                }
            }
        }
        return optimizedScanString;
    }

    // return the keystrokes for a word, given a letter-to-key mapping
    public static String getKeystrokes(String word, int[] letterPosition)
    {
        String keystrokes = "";
        for (int i = 0; i < word.length(); ++i)
            keystrokes += "" + (letterPosition[word.toLowerCase().charAt(i) - 'a'] + 1);
        return keystrokes;
    }

    // return a distance matrix give the scan distance between any two letters,
    // given a letter-to-key mapping
    private static int[][] getDistanceMatrix(int[] letterPosition)
    {
        String[] keyboard = getKeyboard(letterPosition);
        int dist[][] = new int[26][26];
        for (int i = 0; i < 26; ++i)
        {
            for (int j = 0; j < 26; ++j)
            {
                int p1 = letterPosition[i];
                int p2 = letterPosition[j];
                int d = p2 - p1;
                if (d > 0)
                    dist[i][j] = d;
                else
                    dist[i][j] = d + keyboard.length + 1;
            }
        }
        return dist;
    }

    // return a string array containing the letters on each key of virtual keyboard
    private static String[] getKeyboard(int[] letterPosition)
    {
        String[] keyboard
                = new String[letterPosition[25] + 1]; // 'z' position tells all
        for (int i = 0; i < keyboard.length; ++i)
            keyboard[i] = "";
        final String LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
        for (int i = 0; i < 26; ++i)
            keyboard[letterPosition[i]] += LETTERS.charAt(i);
        return keyboard;
    }

    // return an int array containing the position of each letter on the keyboard
    public static int[] getLetterPosition(EncodedWord[] ew)
    {
        int[] letterPosition = new int[26];
        for (int i = 0; i < 26; ++i)
        {
            for (EncodedWord encodedWord : ew)
            {
                char c = (char)(i + 'a');
                int idx = encodedWord.getWord().indexOf(c);
                if (idx >= 0)
                {
                    letterPosition[i]
                            = Integer.parseInt(encodedWord.getCode().substring(idx, idx + 1)) - 1;
                    break;
                }
            }
        }
        return letterPosition;
    }

    // return a query with candidate words
    private static String[] getCandidates(String keystrokesArg, EncodedWord[] ewArg)
    {
        String[] s = EncodedWord.getQuery(keystrokesArg, ewArg);
        if (s == null)
            return null;

        int size = s.length > 100 ? 100 : s.length;
        return Arrays.copyOf(s, size);
    }

    // return the number of possible MLPK opportunities for a word
    public static int getMLPKOpportunities(String word, EncodedWord[] ew)
    {
        int[] letterPosition = getLetterPosition(ew);
        String keystrokes = getKeystrokes(word, letterPosition);
        boolean doubleUp = false;
        String reducedKeystrokes = "";
        reducedKeystrokes += keystrokes.charAt(0);

        int opportunities = 0;
        for (int j = 1; j < keystrokes.length(); ++j)
        {
            char lastKeystroke = reducedKeystrokes.charAt(reducedKeystrokes.length() - 1);
            char c = keystrokes.charAt(j);
            if (lastKeystroke == c && !doubleUp)
            {
                doubleUp = true;
                ++opportunities;
            } else
            {
                doubleUp = false;
                reducedKeystrokes += c;
            }
        }
        return opportunities;
    }

    // return 1 if there is a OW opportunity, 0 otherwise
    public static int getOWOpportunities(String word, EncodedWord[] ew)
    {
        int opportunities = 0;

        String scanString;
        int[] letterPosition;
        String[] keyboard;
        letterPosition = getLetterPosition(ew);
        keyboard = getKeyboard(letterPosition);
        String keystrokes = getKeystrokes(word, letterPosition);

        // build an array showing position of candidate word in list for
        // each word stem (e.g., "hello" --> h, he, hel, hell, hello)
        int[] candidateIndex = new int[word.length()];
        for (int j = 0; j < word.length(); ++j)
        {
            String keystrokeStem = keystrokes.substring(0, j + 1);
            String[] tryQuery = getCandidates(keystrokeStem, ew);
            if (tryQuery == null)
                return 0;
            candidateIndex[j] = -1; // assume word not in candidate list
            for (int k = 0; k < tryQuery.length; ++k)
                if (word.equals(tryQuery[k]))
                {
                    candidateIndex[j] = k;
                    break;
                }
        }

        // use MLPK scan string as the starting point
        scanString = getMLPKScanSequence(word, ew);
        String optimizedScanString;
        String wordStem = "";
        for (int i = 0; scanString != null && i < scanString.length(); ++i)
        {
            char c = scanString.charAt(i);
            if (Character.isLowerCase(c))
            {
                String scanStem = scanString.substring(0, i + 1);
                wordStem += c;

                // check if doubled-up letter (but not if already at end of word!)
                if (wordStem.length() < word.length())
                {
                    char next = word.charAt(wordStem.length());
                    if (letterPosition[c - 'a'] == letterPosition[next - 'a'])
                        wordStem += next;
                }

                int p1 = scanStem.length() + (keyboard.length - letterPosition[c - 'a']);
                int p2 = candidateIndex[wordStem.length() - 1];
                if (p2 < 0) p2 = 99999;
                if ((p1 + p2) < scanString.length())
                {
                    optimizedScanString = scanStem;
                    for (int j = 0; j < keyboard.length - letterPosition[c - 'a'] - 1; ++j)
                        optimizedScanString += ".";
                    optimizedScanString += "S";
                    for (int j = 0; j < candidateIndex[wordStem.length() - 1]; ++j)
                        optimizedScanString += ".";
                    optimizedScanString += "W";

                    // the moment of truth...
                    if (optimizedScanString.length() < scanString.length())
                        opportunities = 1;
                    break;
                }
            }
        }
        return opportunities;
    }
}